from ufrc.sbatch import SBatchHeaders, SBatchFile

EXPECT_SBATCH_HEADERS = """
#SBATCH --job-name=my_job
#SBATCH --mail-user=derobertisna@ufl.edu
#SBATCH --mem=1000mb
#SBATCH --ntasks=1
#SBATCH --cpus-per-task=1
#SBATCH --mail-type=END,FAIL
#SBATCH --time=14-0:00
#SBATCH --output=%j.log
""".strip()

EXPECT_SBATCH_FILE = f"""
#!/bin/bash
{EXPECT_SBATCH_HEADERS}
module load python
ls -l
""".strip()


def test_sbatch_headers():
    sbatch = SBatchHeaders(job_name="my_job", email="derobertisna@ufl.edu")
    output = sbatch.header_str
    assert output == EXPECT_SBATCH_HEADERS


def test_sbatch_headers_with_array():
    sbatch = SBatchHeaders(job_name="my_job", email="derobertisna@ufl.edu", array="1-5")
    output = sbatch.header_str
    assert output == EXPECT_SBATCH_HEADERS + "\n#SBATCH --array=1-5"


def test_sbatch_file():
    headers = SBatchHeaders(job_name="my_job", email="derobertisna@ufl.edu")
    sbatch = SBatchFile(commands=["module load python", "ls -l"], headers=headers)
    output = sbatch.contents
    assert output == EXPECT_SBATCH_FILE



from settings import Settings, SETTINGS
from ufrc.main import UFRC
from ufrc.sbatch import SBatchHeaders, SBatchFile


def put_sbatch(settings: Settings):
    ufrc = UFRC()
    ufrc.connect(settings.username, settings.password)
    headers = SBatchHeaders(job_name="my_job", email="derobertisna@ufl.edu")
    sbatch = SBatchFile(commands=["module load python", "ls -l"], headers=headers)
    ufrc.put_sbatch(sbatch)


if __name__ == "__main__":
    put_sbatch(SETTINGS)



from settings import Settings, SETTINGS
from ufrc.main import UFRC


def put_file(settings: Settings):
    ufrc = UFRC()
    ufrc.connect(settings.username, settings.password)
    ufrc.put("codecov.yml")


if __name__ == "__main__":
    put_file(SETTINGS)



import os
import tempfile
from pathlib import Path
from typing import Final, Optional, Sequence, Union

import scp
from scp import SCPClient

from ufrc.exc import NoUFRCConnectionException
from ufrc.sbatch import SBatchFile
from ufrc.ssh.client import SSHClient
from ufrc.ssh.response import SSHResponse

SERVER: Final[str] = "hpg.rc.ufl.edu"


class UFRC:
    def __init__(self, server: str = SERVER):
        self.server = server

        self._ssh_client: Optional[SSHClient] = None
        self._scp_client: Optional[SCPClient] = None

    def connect(self, username: str, password: str):
        ssh = SSHClient()
        ssh.load_system_host_keys()
        ssh.connect(self.server, username=username, password=password)
        self._ssh_client = ssh
        self._scp_client = SCPClient(ssh.get_transport())

    def disconnect(self):
        self._ssh_client.close()
        del self._ssh_client
        del self._scp_client
        self._ssh_client = None
        self._scp_client = None

    @property
    def is_connected(self) -> bool:
        return self._ssh_client is not None

    def run(self, command: str) -> SSHResponse:
        if self._ssh_client is None:
            raise NoUFRCConnectionException(
                "Use .connect(username, password) before .run"
            )

        _, ssh_stdout, ssh_stderr = self._ssh_client.exec_command(command)
        return SSHResponse(
            stdout=ssh_stdout.read().decode(),
            stderr=ssh_stderr.read().decode(),
        )

    def put(
        self,
        files: Union[str, Sequence[str]],
        to_path: str = ".",
        recursive=False,
        preserve_times=False,
    ):
        if self._scp_client is None:
            raise NoUFRCConnectionException(
                "Use .connect(username, password) before .put"
            )

        self._scp_client.put(
            files, to_path, recursive=recursive, preserve_times=preserve_times
        )

    def put_sbatch(
        self, sbatch: SBatchFile, file_name: str = "my_job.sbatch", to_path: str = "."
    ):
        with tempfile.TemporaryDirectory() as temp_dir:
            out_path = Path(temp_dir) / file_name
            out_path.write_text(sbatch.contents)
            self.put(str(out_path), to_path=to_path)

    def get(
        self,
        remote_path: str,
        local_path: str = "",
        recursive=False,
        preserve_times=False,
    ):
        if self._scp_client is None:
            raise NoUFRCConnectionException(
                "Use .connect(username, password) before .get"
            )

        self._scp_client.get(
            remote_path, local_path, recursive=recursive, preserve_times=preserve_times
        )

    # TODO: run python code directly



from enum import Enum
from typing import List, Union, Optional

from pydantic import BaseModel, Field


class MailType(str, Enum):
    NONE = "NONE"
    BEGIN = "BEGIN"
    END = "END"
    FAIL = "FAIL"
    ALL = "ALL"


class SBatchHeaders(BaseModel):
    job_name: str
    email: str
    memory_mb: int = 1000
    n_tasks: int = 1
    cpus_per_task: int = 1
    mail_types: List[MailType] = Field(
        default_factory=lambda: [MailType.END, MailType.FAIL]
    )
    output: str = "%j.log"
    time: str = "14-0:00"
    array: Optional[str] = None

    @property
    def header_str(self) -> str:
        headers = [
            _sbatch_line("job-name", self.job_name),
            _sbatch_line("mail-user", self.email),
            _sbatch_line("mem", f"{self.memory_mb}mb"),
            _sbatch_line("ntasks", self.n_tasks),
            _sbatch_line("cpus-per-task", self.cpus_per_task),
            _sbatch_line("mail-type", self.mail_types),  # type: ignore
            _sbatch_line("time", self.time),
            _sbatch_line("output", self.output),
        ]
        if self.array is not None:
            headers.append(_sbatch_line("array", self.array))
        return "\n".join(headers)


SBatchPrimitive = Union[str, int, MailType]
SBatchValue = Union[SBatchPrimitive, List[SBatchPrimitive]]


def _sbatch_line(flag_name: str, value: SBatchValue) -> str:
    return f"#SBATCH --{flag_name}={_to_sbatch_value_str(value)}"


def _to_sbatch_value_str(value: SBatchValue) -> str:
    if isinstance(value, list):
        return ",".join(_sbatch_primitive_to_str(v) for v in value)
    return _sbatch_primitive_to_str(value)


def _sbatch_primitive_to_str(value: SBatchPrimitive) -> str:
    if isinstance(value, Enum):
        return str(value.value)
    return str(value)


class SBatchFile(BaseModel):
    commands: List[str]
    headers: SBatchHeaders
    shebang: str = "#!/bin/bash"

    @property
    def contents(self) -> str:
        outputs = [self.shebang, self.headers.header_str, *self.commands]
        return "\n".join(outputs)



from settings import Settings, SETTINGS
from ufrc.main import UFRC


def get_file(settings: Settings):
    ufrc = UFRC()
    ufrc.connect(settings.username, settings.password)
    ufrc.get("temp.txt")


if __name__ == "__main__":
    get_file(SETTINGS)



import pytest

from ufrc.exc import NoUFRCConnectionException
from ufrc.main import UFRC

# TODO: add automated tests by mocking UFRC SSH session


def test_cannot_run_when_not_connected():
    rc = UFRC()
    with pytest.raises(NoUFRCConnectionException):
        rc.run("ls")



from settings import Settings, SETTINGS
from ufrc.main import UFRC


def run_ls(settings: Settings):
    ufrc = UFRC()
    ufrc.connect(settings.username, settings.password)
    output = ufrc.run("ls")
    print(output)


if __name__ == "__main__":
    run_ls(SETTINGS)



import paramiko


class SSHClient(paramiko.SSHClient):
    def handler(self, title, instructions, prompt_list):
        answers = []

        for prompt_, _ in prompt_list:
            prompt = prompt_.strip().lower()
            print("'" + prompt + "'", prompt.startswith("Duo two-factor login"))
            if prompt.startswith("password"):
                answers.append(self.password)
            # TODO: Add ability to pass option 2 for a phone call
            elif prompt.startswith("duo two-factor login"):
                answers.append("1")
            else:
                raise ValueError("Unknown prompt: {}".format(prompt_))

        return answers

    def auth_interactive(self, username, handler):
        self._transport.auth_interactive(username, handler)

    def _auth(self, username, password, pkey, *args):
        self.password = password
        return self.auth_interactive(username, self.handler)
